# Calculates directional slope and expresses it as a percentage
# Written by Sterling Quinn

# Some logic adapted from code posted by user FelixIP on StackExchange
# https://gis.stackexchange.com/questions/136715/getting-cell-value-along-flow-direction-using-arcpy

# This code is made available under The MIT License

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import arcpy, os, traceback, sys, numpy, math
from arcpy import env
from arcpy.sa import *
env.overwriteOutput = True

# Paths to input flow direction raster and elevation dem
fdir=r'c:\data\snoq.gdb\fdir'
elev =r'c:\data\snoq.gdb\demfill'

# Path to output
output =r'c:\data\snoq.gdb\output'

try:
    def showPyMessage():
        arcpy.AddMessage(str(time.ctime()) + " - " + message)

#   Convert rasters to numpy arrays
    fdirArray = arcpy.RasterToNumPyArray(fdir, "", "", "", -9999)
    elevArray = arcpy.RasterToNumPyArray(elev, "", "", "", -9999)

    nRows,nCols=fdirArray.shape
    cellsTotal=nCols*nRows
    d=arcpy.Describe(fdir)
    sr = d.spatialReference
    origin=d.extent.lowerLeft
    cSize=arcpy.Raster(fdir).meanCellHeight

#   Constants to help find cell neighbor
    fDirs=(1,2,4,8,16,32,64,128)
    rookDirs = (1,4,16,64)
    diagDirs = (2,8,32,128)
    dCol=(1,  1,  0, -1, -1,-1, 0,1)
    dRow=(0,  1,  1,  1,  0, -1, -1,-1)
    
#   Loop through raster cells
    arcpy.SetProgressor("step", "", 0, cellsTotal)
    blankArray = numpy.empty((nRows,nCols))
    for nRow in range (nRows):
        for nCol in range (nCols):
            arcpy.SetProgressorPosition()

            # Get value of DEM pixel
            elevPixel=elevArray[nRow, nCol]
            if elevPixel == -9999: continue

            # Get value of corresponding flow direction pixel
            fdirPixel=fdirArray[nRow,nCol]
            i = fDirs.index(fdirPixel)

            # Get the column of the cell to compare
            nC = nCol + dCol[i]
            if nC < 0 or nC == nCols: continue

            # Get the row of the cell to compare
            nR = nRow + dRow[i]
            if nR < 0 or nR == nRows: continue

            # Get the elevation pixel value of the cell to compare
            dirElevPixel = elevArray[nR, nC]

            # Calculate the difference between elevation pixels
            elevPixelDiff = elevPixel - dirElevPixel

            maxSlope = -9999

            # If cell is in diagonal direction, divide the difference by the orthogonal size
            if fdirPixel in diagDirs:
                maxSlope = ((elevPixelDiff * 1.0) / (math.sqrt(2 * (cSize * cSize))))
            # If cell is in rook direction, divide the difference by the cell size
            elif fdirPixel in rookDirs:
                maxSlope = ((elevPixelDiff * 1.0) / cSize)
            else:
                maxSlope = -9999

            # Write the output cell value to the output array
            blankArray[nRow,nCol] = maxSlope * 100


    # Convert the output array to a raster
    myRaster = arcpy.NumPyArrayToRaster(blankArray,origin,cSize,cSize)
    arcpy.DefineProjection_management(myRaster, sr)
    myRaster.save(output)
    del fdirArray,elevArray,blankArray

except:
    message = "\n*** PYTHON ERRORS *** "; showPyMessage()
    message = "Python Traceback Info: " + traceback.format_tb(sys.exc_info()[2])[0]; showPyMessage()
    message = "Python Error Info: " +  str(sys.exc_type)+ ": " + str(sys.exc_value) + "\n"; showPyMessage()
